# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# kdb - the Komodo debugger!

# When run as a top-level script, this program debugs another Python program,
# communicating with the remote debugger UI.

# IMPORTANT: remote debugging code (remote end) should not rely on Python 2.0
# or people will not be able to debug programs on servers running old versions.


# When imported, it functions as the listener in the debugger UI.
import os
import sys
import bdb
import string
import threading, thread
import types
import traceback
import base64
import whrandom
import time
import korpc
import socket
import urlparse
import urllib

# I dont want this code to depend on xpcom!!
RESUME_STOP = 0 # session terminated.
RESUME_STEP_IN = 1 # step into things.
RESUME_STEP_OVER = 2 # step over current thing
RESUME_STEP_OUT = 3 # step out of current thing.
RESUME_GO = 4 # go for it.

BREAK_REASON_BREAK = 1

# Some state constants to work around 2 limitations in pdb:
# * When at the bottom frame, all operations act like "step in"
# * Starting a debug session always stops at the first line in
#    the program - there is no way to start "running" so the first
#    stop is at the first breakpoint.
BOTFRAME_STEP = 0 # When next at the bottom frame, we will break
BOTFRAME_CONTINUE = 1 # When next at the bottom frame, we just ignore it.

AllStringTypes = [types.StringType]
startupDir = os.path.abspath(os.curdir)

# types.UnicodeType is a 1.6 feature
if hasattr(types, "UnicodeType"):
    AllStringTypes.append(types.UnicodeType)

def url2pathname( url ):
    decomposedURL = urlparse.urlparse( url, 'file:///' )

    path = decomposedURL[2]
    # The path will begin with a superfluous '/' in this case
    if ( string.lower(string.strip(url))[0:8] == 'file:///' and string.find(sys.platform, 'win') == 0):
        path = path[1:]
        
    return urllib.url2pathname( path )

def pathname2url( pathname ):
    return 'file:' + urllib.pathname2url( os.path.abspath( pathname ) )

def makeCookie():
    cooklen=8
    if os.path.exists("/dev/random"):
        rand=open("/dev/random", "r")
        data=rand.read(cooklen)
    else:
        generator=whrandom.whrandom()
        generator.seed(int(time.time())%256, sys.getrefcount(None)%256,
                 sys.getrefcount(0)%256)
        data=""
        for i in range(cooklen):
            data=data+chr(generator.randint(0,127))
    return string.strip(base64.encodestring(data))

def _frame_repr(frame,name=None):
    if name is None: name = ""
    if frame:
        if frame.f_code and frame.f_code.co_filename:
            fname = os.path.split(frame.f_code.co_filename)[1]
        else:
            fname = "??"
        return "Frame %s: %s,%d (%r)" % (name, fname, frame.f_lineno, frame)
    else:
        return "Frame %s: None" % (name,)

def _format_exception_only():
    try:
        # Assume exception can be formatted as a list of with single element, with a trailing \n.
        return traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])[0][:-1]
    except IndexError:
        return str(sys.exc_info()[0]) + ": " + str(sys.exc_info()[1])

def _get_object_type_string(value):
    try:
        return type(value).__name__
    except AttributeError:
        return str(type(value))

def _safe_apply(what_failed, func, *args):
    try:
        return apply(func, args)
    except:
        return "%s failed - %s: %s" % (what_failed, sys.exc_info()[0], sys.exc_info()[1])

def _safe_index(what_failed, ob, index):
    try:
        return ob[index]
    except:
        return "%s[%s] failed - %s: %s" % (what_failed, index, sys.exc_info()[0], sys.exc_info()[1])

class Kdb(bdb.Bdb):
    def __init__(self, requester):
        self.requester = requester
        # Default is to step at the bottom frame.
        self.botframeBehaviour = BOTFRAME_STEP
        bdb.Bdb.__init__(self)
        self.forget()

    def __del__(self):
        # XXX - there is a cycle/leak - this object never dies :-(
        print "Kdb", self, "dieing"

    def canonic(self, fname):
        # One Python 1.5.2 Bdb.bdb wierdism was that it doesn't define canonic if a subclass does!
        # But it's a no-op so we can ignore it anyway...
        try:
            fullname = os.path.normcase(bdb.Bdb.canonic(self, fname))
        except AttributeError:
            fullname = os.path.normcase(os.path.abspath(fname))
        # Work around a problem where some filename in Python code
        # objects are relative.
        if not os.path.exists(fullname):
            fullname = None

            # Try looking through the module search path
            for dirname in sys.path:
                testname = os.path.join(dirname, fname)
                if os.path.exists(testname):
                   fullname = os.path.abspath(testname)
                   break
        return fullname

    
    def setup(self, f, t):
        self.forget()
        self.stack, self.curindex = self.get_stack(f, t)
        self.stack.reverse()

    def forget(self):
        self.stack = []
        self.curindex = 0

    def set_break(self, filename, lineno, temporary=0, cond = None):
        # A clone of bdb, except we remember the result.
        filename = self.canonic(filename)
        import linecache # Import as late as possible
        line = linecache.getline(filename, lineno)
        if not line:
            return None
        if not self.breaks.has_key(filename):
            self.breaks[filename] = []
        list = self.breaks[filename]
        if not lineno in list:
            list.append(lineno)
        return bdb.Breakpoint(filename, lineno, temporary, cond)

    def interaction(self, frame, tb):
        self.setup(frame, tb)
        try:
            return self.requester._serveRequests()
        finally:
            self.forget()

    def user_line(self, frame):
        # Python has this thing about line zero :-)
#        sys.__stderr__.write("user_line at " + _frame_repr(frame) + "with locals" + str(frame.f_locals) + "\n");
        if frame.f_lineno == 0:
            return
        if frame.f_locals.has_key("__no_komodo__"):
            # Internal Komodo function - get outa the entire function!
            self.set_return(frame)
            return
        rc = self.interaction(frame, None)
        if rc == RESUME_STOP:
            self.set_quit()
        elif rc == RESUME_STEP_IN:
            self.set_step()
        elif rc == RESUME_STEP_OVER:
            self.set_next(frame)
        elif rc == RESUME_STEP_OUT:
            self.set_return(frame)
        elif rc == RESUME_GO:
            self.set_continue()
        else:
            print "Unknown resume state", rc

    # Override the dbd set_* functions to manage our bottom frame fix.
    def set_quit(self):
        self.botframeBehaviour = BOTFRAME_CONTINUE
        bdb.Bdb.set_quit(self)
    def set_step(self):
        self.botframeBehaviour = BOTFRAME_STEP
        bdb.Bdb.set_step(self)
    def set_continue(self):
        self.botframeBehaviour = BOTFRAME_CONTINUE
        bdb.Bdb.set_continue(self)
    def set_next(self, frame):
        # A "step over" operation - if we are at the bottom frame,
        # then we need to be in step mode.
        if frame is self.botframe:
            self.botframeBehaviour = BOTFRAME_STEP
        bdb.Bdb.set_next(self, frame)
    def set_return(self, frame):
        # A "step out" operation - if the frame we are returning to is
        # the bottom frame, then we need to be in step mode.
        if frame.f_back is self.botframe:
            self.botframeBehaviour = BOTFRAME_STEP
        elif frame is self.botframe:
            # At the bot-frame and stepping out - like a "run"
            self.botframeBehaviour = BOTFRAME_CONTINUE
        bdb.Bdb.set_return(self, frame)

    def stop_here(self, frame):
        if frame is self.botframe and \
           self.botframeBehaviour == BOTFRAME_CONTINUE and \
           not self.break_here(frame):
            self.set_continue()
            return 0
        if frame is self.botframe and self.botframeBehaviour == BOTFRAME_STEP:
            self.set_step()
            return 1
        return bdb.Bdb.stop_here(self, frame)


class SocketHandler:
    def __init__(self, sock, also_to = None):
        self.file = sock.makefile()
        self.file_also_to = also_to
        self.reported_error = 0
    def _safe_apply_(self, fn_name, do_flush, *args):
    
        main_func = getattr(self.file, fn_name)
        also_func = getattr(self.file_also_to, fn_name)
        try:
            try:
                apply(main_func, args)

                # flush the socket on write so that Komodo displays
                # output as we go
                if do_flush:
                    self.file.flush()

                self.reported_error = 0
            except:
                if not self.reported_error and self.file_also_to is not None:
                    self.file_also_to.write("*** Error redirecting output to Komodo - %s: %s\n" % (sys.exc_info()[0], sys.exc_info()[1]) )
                    self.reported_error = 1
        finally:
            # Dont catch exceptions writing to the original file - if that
            # raises an exception, it still would even if we arent here, so
            # the caller _should_ see it!
            apply(also_func, args)
        
    def write(self, what, __no_komodo__ = None):
        self._safe_apply_("write", 1, what)
    def flush(self, __no_komodo__ = None):
        self._safe_apply_("flush", 0)
    def close(self, __no_komodo__ = None):
        self._safe_apply_("close", 0)
    def readline(self, __no_komodo__ = None):
        return self.file.readline()

class Requester:
    def __init__(self):
        self.__original_stdout = None
        self.__original_stderr = None
        self.__original_stdin = None

    def __del__(self):
        # XXX - there is a cycle/leak - this object never dies :-(
        print self, "dieing"

    def brk(self):
        self._do_break()
    def _do_break(self):
        if not hasattr(self, "kdb"):
            self.kdb = Kdb(self)
        self.kdb.set_trace()

    def start(self, debug_args):
        # This blocks until we are finished.
        self._do_start(debug_args)
        return 1
    def _do_start(self, debug_args):
        self.kdb = Kdb(self)

        # Run our initial command-loop to tell the debugger we are alive,
        # and get the initial start request
        # NOTE: We coould now support pushing the program
        # and args down the command channel.
        step_type = self._serveRequests()
        if step_type == RESUME_GO:
            self.kdb.botframeBehaviour = BOTFRAME_CONTINUE

        # Ready to roll - do it.
        code = open(debug_args[0]).read()
        
        # python expects lines to end with '\n' only so kill the '\r's
        code = string.replace(code, '\r\n', '\n')
        code = string.replace(code, '\r', '\n')

        # there's also an issue with files that don't end with newlines, so tag on a few
        code = code + '\n\n\n'
        
        code_ob = compile(code, debug_args[0], "exec")
        old_argv = getattr(sys, 'argv', [])
        sys.argv = debug_args
        locals = {}
        import __main__
        for key in __main__.__dict__.keys():
            if string.find(key, "__") == 0:
                locals[key]=__main__.__dict__[key]
        try:
            self.kdb.run(code_ob, locals, locals)
            self.kdb = None
        except:
            traceback.print_exc()
        
        sys.argv = old_argv
    def redirectStdout(self, port, **otherParams ):
        if self.__original_stdout == None:
            self.__original_stdout = sys.stdout
            
        (hostname,_junk_port) = self.request.getpeername()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((hostname,port))
        sys.stdout = SocketHandler(sock, self.__original_stdout)
        return {}
    def redirectStderr(self, port, **otherParams):
        if self.__original_stderr == None:
            self.__original_stderr = sys.stderr
        
        (hostname,_junk_port) = self.request.getpeername()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((hostname,port))
        sys.stderr = SocketHandler(sock, self.__original_stderr)
        return {}
    def redirectStdin(self, port, **otherParams):
        if self.__original_stdin == None:
            self.__original_stdin = sys.stdin
        
        (hostname,_junk_port) = self.request.getpeername()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((hostname,port))
        sys.stdin = SocketHandler(sock, self.__original_stdin)
        return {}
    def hello(self, **otherParams ):
        return { 'status' : 'hello' }
    def resumeExecution(self, request, **otherParams ):
        self.continue_request = request
    def getNumFrames(self, **otherParams):
        return { 'numFrames' : len(self.kdb.stack) }
    def getFrames(self, **otherParams ):
        ret = []
        for item in self.kdb.stack:
            frame, lineno = item
            where = frame.f_code.co_name or ''
            codeFileName = frame.f_code.co_filename

            if os.path.isfile(codeFileName):
                filename = codeFileName
            else:
                # OK, what probably happened is that the .py[c,o] file was relocated and no longer points to
                # a correct source file. We will try to find the source file in the same directory as the
                # .py[c,o] file.
                if not frame.f_globals.has_key('__file__'):
                    filename = codeFileName # This actually seems possible based on import.c
                else:
                    moduleFileName = frame.f_globals['__file__']
                    if os.path.isabs(moduleFileName):
                        filename = os.path.join(os.path.dirname(moduleFileName), os.path.basename(codeFileName))
                    else:
                        # Assume that all module imports were relative the the current directory at kdb startup.
                        # This should work most of the time. Heck, it might not even be necessary because, due
                        # to the compile, all paths seem to be absolute.
                        filename = os.path.join(startupDir, os.path.dirname(moduleFileName), os.path.basename(codeFileName))
                    
            ret.append( { 'file' : pathname2url( filename ), 'lineno' : lineno, 'where' : where} )
        return { 'frames' : ret }
    def getSourceCode(self, filename, **otherParams):
#        sys.__stderr__.write("getSourceCode for " + str(filename) + "\n")
        try: f = open(url2pathname( filename ))
        except IOError: return { 'source' :"# Source code unavailable" }
        else: return { 'source': f.read() }
    def _getObject(self, frame_index, expr):
        frame, lineno = self.kdb.stack[frame_index]
        try:
            value = eval(expr, frame.f_globals, frame.f_locals)
            typ = _get_object_type_string(value)
        except:
            value = _format_exception_only()
            typ = "Error"

        return value, typ

    def _getContextObject(self, context, frame_index, expr ):
        frame, lineno = self.kdb.stack[frame_index]
        if context == 'locals':
            bindings = frame.f_locals
        elif context == 'globals':
            bindings = frame.f_globals

        try:
            value = eval(expr, bindings)
            typ = _get_object_type_string(value)
        except:
            value = _format_exception_only()
            typ = "Error"

        return value, typ
    
    def getDebugPropertyInfo(self, frame_index, expr, radix, **otherParams ):
        value, typ = self._getObject(frame_index, expr)
        valueString, has_children = self._getDebugPropertyInfo(value, radix)

        return { 'propertyInfo' : { 'name' : { 'name' : expr, 'fullName' : expr },
                                    'value' : valueString, 'type' : typ, 'hasChildren' : has_children } }

    def _getDebugPropertyInfo(self, value, radix):
        valueString = stringify(value)
        # checking for properties should catch the vast majority of cases!
        if type(value) in AllStringTypes: 
           has_children = 0
        else:
            non_methods = []
            for child in dir(value):
                if type(child) != type(types.MethodType):
                    non_methods.append(child)

            has_children = len(non_methods) != 0

            if not has_children: # last resort -- try hairy algorithms
                has_children = len(self._getChildrenItems(value))!=0
        return valueString, has_children

    def setDebugProperty(self, frame_index, lhs, rhs, radix, context = None, **otherParams):
        # XXX - to do - don't ignore radix!
        # Returns an error message or an empty string.
        frame, lineno = self.kdb.stack[frame_index]
        """
        if context == 'globals':
            try:
                # It was decided that the RHS should be evaluated in the local context
                rhs_ob = eval(rhs, frame.f_globals ,frame.f_locals)
            except:
                return { 'setResultString' : _format_exception_only() }
                   
            frame.f_globals [lhs] = rhs_ob
        """
        # If the frame locals does not have a variable of this
        # name, but the globals does, then almost certainly the
        # global is what they want to update.  "eval" doesnt handle this.            
        if (context == 'globals') or (not frame.f_locals.has_key(lhs) and frame.f_globals.has_key(lhs)):
            # Eval the rhs first.
            try:
                rhs_ob = eval(rhs, frame.f_globals, frame.f_locals)
            except:
                return { 'setResultString' : _format_exception_only() }
            frame.f_globals[lhs] = rhs_ob

        else:
            # Hack an assignment together.
            # XXX This doesn't always work because of LOAD_FAST. Either we need to work around or
            # change this code to generate an error message if the assignment fails.
            
            expr = lhs + " = " + rhs
            expr = str(expr)
            #print 'exec %s in %s, %s' % (repr(expr), repr(frame.f_globals), repr(frame.f_locals))
            try:
                exec expr in frame.f_globals, frame.f_locals
                #print "frame.f_globals:", frame.f_globals
                #print "frame.f_locals:", frame.f_locals
            except:
                return { 'setResultString' : _format_exception_only() }
        return { 'setResultString' : '' } # Life is good!
    
    def getChildrenInfos(self, frame_index, parent_expr, radix, context = None, **otherParams):
        if context:
            value, typ = self._getContextObject(context, frame_index, parent_expr)
        else:
            value, typ = self._getObject(frame_index, parent_expr)
            
        children = self._getChildrenItems(value)
        def mycmp(i,j): return cmp(string.lower(i[0]), string.lower(j[0]))
        children.sort(mycmp)
        ret = []
        for (child_name, child_item) in children:
            valueString, has_children = self._getDebugPropertyInfo(child_item, radix)

            ret.append( { 'name': { 'name' :  child_name, 'fullName' : parent_expr + child_name },
                        'value' : valueString,
                        'type' : _get_object_type_string(child_item),
                        'hasChildren' : has_children } )
            
        return { 'propertyInfo' : ret }

    def getChildrenNames(self, frame_index, parent_expr):
        # This function should not be necessary
        raise NotImplementedError
        value, typ = self._getObject(frame_index, parent_expr)
        children = self._getChildrenItems(value)

        result = []
        for item in children:
            result.append(item[0])
            
        return result

    def _getChildrenItems(self, value):
        children = []
        for childStr in dir(value):
            if childStr == '__builtins__':
                continue
            try:
                child = getattr(value,childStr)
            except:
                children.append((".%s" % (childStr), "Getting attribute failed"))
            else:
                if type(child) not in [types.BuiltinMethodType]:
                    children.append((".%s"% (childStr), _safe_apply("Getting attribute failed", getattr, value, childStr)))
                                
        if hasattr(value, "__class__"):
            children.append((".__class__", _safe_apply("Getting class failed!", getattr, value, "__class__")))

        try:
             # repr introduces quotes if the object is already a string
             # (but the object may be one being debugged, and its repr
             # may fail!  This wouldn't be good, as it looks like the debugger failed!
             # XXX Shouldn't return all of really big mappings but how do we know what part the
             # user is interested in?
            try:
                child_keys = value.keys()
            except:
                child_keys = None
            if child_keys is not None:
                for child in child_keys:
                    children.append(("[%s]"%(_safe_apply("repr", repr, child)),_safe_index("child", value, child))) 
        except:
            # The users sequence object returned by keys is raising exceptions, so let's just give up 
            pass

        # value doesn't seem to be a mappying-like thing, let's see if it is a sequence-like thing
        if child_keys is None:
            try:
                # only return 10 items of an array-like thing....
                for child in range(min(len(value),10)):
                    child_item = _safe_index("object", value, child)
                    children.append(("[%d]"%child, child_item))
            except:
                pass
            
        return children
    
    def getContext(self, context, frame_index, radix, **otherParams ):
        has_children = 0
        frame, lineno = self.kdb.stack[frame_index]
        if context == 'locals':
            items = frame.f_locals
        elif context == 'globals':
            items = frame.f_globals
        else:
            raise ValueError, "unknown context type"

        ret = []
        names = items.keys()
        def mycmp(i,j): return cmp(string.lower(i), string.lower(j))
        names.sort(mycmp)
        # Remove __builtins__ everywhere!
        if "__builtins__" in names:
            names.remove('__builtins__')

        for name in names:
            item = items[name]
            valueString, has_children = self._getDebugPropertyInfo(item, radix)
            ret.append( { 'name': { 'name' :  name, 'fullName' : name },
                        'value' : valueString,
                        'type' : _get_object_type_string(item),
                        'hasChildren' : has_children } )
        return { 'propertyInfo' : ret }
    def getContextNamesAsString(self, frame_index, **otherParams ):
        return { 'names' : "locals globals" }
    
    def addBreakpoint(self, file, lineno, temporary=0, cond = None, **otherParams ):
        filePath = url2pathname( file )
        if len( cond ) == 0:
            cond = None
            
        bp = self.kdb.set_break(filePath, lineno, temporary, cond)
        if bp is None:
            return { 'breakpointNumber' : -1 }
        
        return { 'breakpointNumber' : bp.number }
    def removeBreakpoint(self, breakpointNumber, **otherParams ):
        self.kdb.clear_bpbynumber(breakpointNumber)
        return {}

    def setBreakpointAttribute(self, no, attr, val):
        bp = bdb.Breakpoint.bpbynumber[no]
        if bp is None:
            raise "no such breakpoint"
        setattr(bp, attr, val)
        return {}


def findIfHasChildren(thing):
    try:
        if dir(thing):
            return 1
        else:
            return 0
    except:
        # Who knows what an object could do as part of dir!
        return 0

def stringify(thing):
    try:
        typ = type(thing)
        if hasattr(thing, '__name__'):
            if hasattr(typ, '__name__'):
                return thing.__name__ + ' (' + typ.__name__ + ')'
            else:
                return thing.__name__ + ' (' + str(typ) + ')'
        else:
            return repr(thing)
    except:
        return "Looking at object failed - %s: %s" % (sys.exc_info()[0], sys.exc_info()[1])

class LocalRequester(Requester):
    def __init__(self, engine):
        self.engine = engine

    def _serveRequests(self):
        return self.engine._runCommandLoop(BREAK_REASON_BREAK)

def usage(why):
    print why
    prog = os.path.basename(sys.argv[0])
    print "Usage:"
    print prog, "program.py [args ...]"
    sys.exit(1)

# Run as a request server!

class CookieRequestHandler(korpc.RequestHandler):
     def __init__(self, cookie, requester, *args):
        self.cookie=cookie
        self.requester=requester
        apply(korpc.RequestHandler.__init__, (self,)+args)
     def call(self, method, params):
        try:
            cookie = params[0]['cookie']
        
            if cookie!=self.cookie:
                raise RuntimeError, "Cookie does not match"
            
            scopedList = string.split( method, '.' )

            if ( len( scopedList ) != 2 ) or ( scopedList[0] != 'Requester' ):
                raise Exception( 'Method call directed to unknown object' )

            method = scopedList[-1]
            
            assert method[0]!= "_", "Method names may not start with _"

            self.requester.request = self.request
            rc = apply(getattr(self.requester, method), (), params[0])
            if rc is None: # this should be changed but just return an empty record for now
                rc = { }
            sys.stdout.flush()
            sys.stderr.flush()
            return rc
        except:
            print "*** Error calling", method, params
            traceback.print_exc()
            print "<== ERROR"
            sys.stdout.flush()
            sys.stderr.flush()
            raise

class RequestHandlerFactory:
    def __init__(self,cookie,requester):
        self.cookie=cookie
        self.requester=requester
    def __call__(self,*args):
        return apply(CookieRequestHandler,(self.cookie,self.requester)+args)

class RemoteRequester(Requester):
    def __init__(self, cookie, port):
        handler_factory=RequestHandlerFactory(cookie,self)
        #self._server = korpc.SocketServer(handler_factory, port)
        self._server = korpc.FastServer(handler_factory, port)
        # logDebug("Debugger is listening on port %s" % self._get_port()
        Requester.__init__(self)

    def start(self, debug_args):
        try:
            return Requester.start(self, debug_args)
        finally:
            if self._server and self._server.socket:
                try:
                    # This should trigger a socket error in the main requester loop
                    # forcing a break with REASON_STOP.
                    self._server.socket.close()
                except:
                    pass
        
    def _get_port(self):
        return self._server.socket.getsockname()[1]

    def _serveRequests(self):
        self.continue_request = None
        while self.continue_request is None:
            try:
                self._server.handle_request()
            except socket.error:
                return RESUME_STOP
        return self.continue_request

def serve(cookie, script_args, machine, port):
    requester = RemoteRequester(cookie,port)
    requester.start(script_args)

def forkAndServe(cookie, script_args, machine, port):
    requester = RemoteRequester(cookie,port)
    t = threading.Thread(target=requester.start, args=(script_args,))
    t.setDaemon(0)
    t.start()
    return requester._get_port()

def ConnectToListener(komodo_hostname, komodo_port):
    if komodo_hostname == None:
        komodo_hostname=os.environ.get("KODEBUG_HOSTNAME", "localhost")
    if komodo_port == None:
        komodo_port=int(os.environ.get("KODEBUG_PORT","9000"))
    this_hostname = socket.gethostbyaddr(socket.gethostname())[2][0]
    port = 0 # zero means "choose dynamically"
    cookie = makeCookie()
    requestor = RemoteRequester(cookie,port)
    connection_port = requestor._get_port()

    # signal the remote end.
    # this is temporary until the remote signaling is finalized
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((komodo_hostname, komodo_port))
    out = "Python:%d:%s" % (connection_port, cookie)
    s.send(out)
    s.close()
    return requestor, connection_port, cookie


# And start the process being debugged.
if __name__=='__main__':
#    import tempfile
#    orig_handles = sys.stdout, sys.stderr
#    log_name = os.path.join(tempfile.gettempdir(), "kdb.log")
#    log_file = sys.stdout = sys.stderr = open(log_name, "w")

    arg_num = 1
    port=8008
    wait = 0
    #print "args", sys.argv

    while arg_num < len(sys.argv):
        arg = sys.argv[arg_num]
        if arg[0] != '-':
            break
        if arg[1] in "Pp":
            port=int(arg[2:])
        elif arg[1] in "Ww":
            wait = 1
        else:
            usage("Invalid arg: " + arg)
        arg_num=arg_num+1
    if arg_num == len(sys.argv):
        usage("No program specified to debug")
    # print "Remote debugger started"
    
    # To simulate sys.path behaviour for the debuggee, we need
    # to add its path to sys.path.
    try:
        sys.path.insert(0, os.path.abspath(os.path.split(sys.argv[arg_num])[0]))
    except os.error:
        print "Could not obtain the path of the script being debugged!  sys.path may not be as expected!"

    requestor, connection_port, cookie = ConnectToListener("localhost", port)
#    print "Connecting to remote requester on port", connection_port

#    requestor = RemoteRequester(cookie, connection_port)
    requestor.start(sys.argv[arg_num:])

    if wait:
        print "Program terminated - press <ENTER> to close"
        raw_input()
